// CELLINDI.CPP

/*****************************************************************
These source files represent significant changes to the original
WorldCom MFC library.  Therefore, these source files may not be
copied except by permission of the author, Joe Willcoxson.  These
source files may not be included in source or compiled form as any
part of a commercial distribution except by permission of the
author.  These files may be used freely for internal corporate use
such as software produced by an IS departement.  These files may
also may be used freely by students for educational purposes.

No warranty or guarrantee of fitness can be made to these files.
*****************************************************************/

#include "stdafx.h"
#include "cellindi.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

/////////////////////////////////////////////////////////////////////////////
// CCellIndices

// CCellIndices is a utility class to help with CBaseGrid and CCellGrid.
// It provides operations to manage cell indices.

int CCellIndices::IsDownRightOf(const CCellIndices& rhs)
{
 	return (row >= rhs.row && col >= rhs.col);
}

int CCellIndices::IsDownLeftOf(const CCellIndices& rhs)
{
 	return (row >= rhs.row && col <= rhs.col);
}

int CCellIndices::IsUpRightOf(const CCellIndices& rhs)
{
 	return (row <= rhs.row && col >= rhs.col);
}

int CCellIndices::IsUpLeftOf(const CCellIndices& rhs)
{
 	return (row <= rhs.row && col <= rhs.col);
}

#ifdef WIN32
CPrintIndicesArray::CPrintIndicesArray()
{
}

#else

CPrintIndicesArray::CPrintIndicesArray()
: m_pArray(0), m_nSize(0), m_nTotalSize(0), m_nStep(10)
{
   m_pArray = 0;
}

CPrintIndicesArray::~CPrintIndicesArray()
{
	RemoveAll();
}

CCellIndices CPrintIndicesArray::operator[](int index)
{
   ASSERT(index < m_nSize);
   return m_pArray[index];
}

int CPrintIndicesArray::Add(const CCellIndices& ci)
{
   if (m_nSize == m_nTotalSize){
      CCellIndices* pNewCI = new CCellIndices[m_nTotalSize + m_nStep];
      memcpy(pNewCI, m_pArray, m_nSize * sizeof(CCellIndices));
      delete m_pArray;
      m_pArray = pNewCI;
      m_nTotalSize += m_nStep;
   }
   m_pArray[m_nSize] = ci;
   return m_nSize++;
}

void CPrintIndicesArray::RemoveAll()
{
   if (m_nTotalSize != 0){
	   m_nTotalSize = m_nSize = 0;
	   delete [] m_pArray;
	   m_pArray = 0;
   }
}

#endif

// Copyright (c) 1997 by Joe Willcoxson and WorldCom.  All rights reserved.
// E-mail:  chinajoe@aol.com
// URL:  http://users.aol.com/chinajoe
// For great long distance service see http://www.wcom.com

